;returns the "tree normal form" of the given formula
;(this is to be called before parsing)
(defun tnf (f)
(car (collapse-ops (negation-nf (basic-connectives f)))))


;rewrites implication, possibly others into basic operators
(defun basic-connectives (formula)
(case (first formula)
 (:imp
      (list :or
            (list :not (basic-connectives (second formula)))
            (basic-connectives (third formula))))
 (:eqv (basic-connectives      
      (list :and
	    (list :imp (second formula) (third  formula))
	    (list :imp (third formula) (second  formula)))))
 ((:and :or)
      (cons (first formula)
                  (mapcar 'basic-connectives (rest formula))))
 ((:exists :forall)
      (list (first formula)
            (second formula)
            (basic-connectives (third formula))))
 (:not (list :not (basic-connectives (second formula))))
 (otherwise formula)
))

;computes the negation normal form of the formula: applies
;DeMorgan's to push negations to the atoms
(defun negation-nf (formula &optional (polarity t))
(case (first formula)
 ((:and :or)
      (cons (if polarity (first formula) (dual (first formula)))
            (mc (x) (negation-nf x polarity) (cdr formula))))
 ((:exists :forall)
      (list (if polarity (first formula) (dual (first formula)))
            (second formula)
            (negation-nf (third formula) polarity)))
 (:not (negation-nf (second formula) (not polarity)))
 ((:t :f) formula)
 (otherwise (list (if polarity :t :f) formula))
))

;when the child has the same operator, lifts its args to the parent
;equivalently, takes the nf wrt some associativity rewrite rule
(defun collapse-ops (formula &optional state param)
(case (first formula)
 ((:and :or)
        (if (eq (first formula) state)
            (apply #'union* 
		   (mc (arg) 
		       (collapse-ops arg (first formula)) 
		       (cdr formula)))
            (list (cons (first formula)
                        (apply #'union* 
			       (mc (arg) 
				   (collapse-ops arg (first formula)) 
				   (cdr formula)))))))
 ((:exists :forall)
   (let ((vars (copy-list (second formula))))
        (if (eq (first formula) state)
            (progn
                  (nconc param vars)
                  (collapse-ops (third formula) state param))
            (list (list (first formula)
                  vars
                  (car (collapse-ops (third formula) (first formula) vars)))
                  ))))
 (otherwise (list formula))
))
