;translates a given theory
(defun launch (lst e)
(let*
((family (mc (f) 
   (case (formula-head f)
     ((:t :f) (let* ((rst (apply #'append (translate f e)))
		     (pos (first rst))
		     (neg (second rst))
		     (pc (make-clause :op :or :lhs nil :rhs (list (theory-root pos))))
		     (nc (make-clause :op :or :lhs (list (theory-root neg)) :rhs nil)))
		(setf (theory-clauses pos) (list pc))
		(setf (theory-clauses neg) (list nc))
		(list pos neg)))
     (:exists (adjoin-clause (first (translate (formula-args f) e)) f e :top))
     (:or (let* ((arg (mc (y) (translate y e) (formula-args f)))
		 (rst (sections (mc (x) (apply 'append x) arg) e)))
	    (adjoin-clause rst f e :top)))
     (:forall (let* ((arg (translate (formula-args f) e))
		     (arg1 (first arg))
		     (arg2 (second arg)))
		(adjoin-clause arg1 f e :top)
;		      (adjoin-clause arg2 f e :neg)
		      ))
     (t (format t "Error in launch!!")))
   lst))
 (res (sections family e)))
res))

;generate all translations of formula f
;which weigh less than e mass
;returns 2 sets of theories:
;where we let f stay positive and
;where we don't (the contrapositive)
(defun translate (f e)  
; Formula -> Mass -> (List Theory)^2
(case (formula-head f)  ;where head is the top-level connective, Formula is actually a datatype
((:t :f) (let* ((  ref (member (formula-head f) (atompred-nodes (formula-atom f)) :key #'car))
	       (posref (cdr (first ref)))
	       (negref (cdr (second ref)))
	       (posn (make-node 
		      :formula f 
		      :pole :pos     
		      :index (node-index posref)))
	       (negn (make-node 
		      :formula f 
		      :pole :neg    
		      :index (node-index negref))))
	   (list (list (make-theory 
			:nodes (list posn)
			:root posn
			:constraints (if *strict-atoms*
					 (list (cons (formula-atom f)
						     (formula-head f))))))
		 (list (make-theory 
			:nodes (list negn)
			:root negn
			:constraints (list (cons (formula-atom f)
						 (flip (formula-head f)))))))))

(:forall
 (let ((rst (translate (formula-args f) e)))
   (list (adjoin-clause (first  rst) f e :pos) 
	 (adjoin-clause (second rst) f e :neg))))

(:exists
 (let ((rst (translate (formula-args f) e)))
   (list (adjoin-clause (first rst) f e :pos) 
	 nil)))  ;never contrapose existentials

(:and 
 (let ((rst (sections (mc (x) (apply 'append x)
			  (mc (y) (translate y e) (formula-args f)) )
		      e)))
   (list (adjoin-clause rst f e :pos)
	 (adjoin-clause rst f e :neg))))

(:or    ;disjuncts can move either way, so we put them together
 (let ((rst (sections (mc (x) (apply 'append x)   
			  (mc (y) (translate y e) (formula-args f)) ) e)))
   (list (adjoin-clause rst f e :pos)  
	 (adjoin-clause rst f e :neg))
))))

;this adds a new clause to each theory in a list,
;doing contraposition when polarity is :NEG
(defun adjoin-clause (lst f max p)   
; (List Theory) -> Node -> Mass -> Pole -> Mass
(let ((n (make-node :formula f :pole p)))
(labels ((build-clause (th)
	   (case (formula-head f)
	     (:forall
	      (make-clause :op (formula-head f)
			   :lhs (case p (:pos (list n)) (:neg (list (theory-root th))) (t nil))
			   :rhs (if (eq p :neg) (list n) (list (theory-root th)))))
	     (:exists
	      (make-clause :op (formula-head f)
			   :lhs (case p (:pos (list n)) (:neg (list (theory-root th))) (t nil))
			   :rhs (if (eq p :neg) (list n) (list (theory-root th)))))
	     (:or
	      (do* ((c  (make-clause :op :or :lhs nil :rhs nil))
		    (ti (theory-aux th) (cdr ti))
		    (r  (car ti) (car ti)))
		   ((null ti)
		    (case p
		      (:pos (push n (clause-lhs c)))
		      (:neg (push n (clause-rhs c)))
		      (t nil))
		    c)
		(if (eq (node-pole r) :pos)
		    (push r (clause-rhs c))
		    (push r (clause-lhs c)))))
	     (t (format t "Get the fuck outta here!!~%"))
	 ))
	 (list-if (x s) (if (eq (node-pole x) s) (list x)))
	 (build-conj (lst)
	   (let* ((args (funcall (make-binary-sort (lambda (x) (eq (node-pole x) :pos))) lst))
		  (pcs (nreverse (car args)))
		  (ncs (nreverse (cdr args))))
;	     (break)
	     (if (eq p :pos)
		 (cons     (make-clause :op :and 
					:lhs (list n)
					:rhs (or pcs (list n)))
		       (mc (x) 
			   (make-clause :op :and
					:lhs (list n x)
					:rhs nil)
			   ncs))
		 (mc (x)   (make-clause :op :and
					:lhs (list-if x :neg)
					:rhs (cons n (list-if x :pos)))
		     (append pcs ncs))
	 )))
	 (build-theory (th)
	   (let* ((c (if (eq (formula-head f) :and)
			 (build-conj (theory-aux th))
			 (build-clause th)))
		  (dm (if (eq (formula-head f) :and)
			  (make-mass :v (apply #'+ (mapcar #'weigh c)))
			  (make-mass :v (weigh c)))))
	     (make-theory              
	         :nodes   (cons n (theory-nodes th))
		 :clauses (funcall (if (listp c) #'append #'cons) c (theory-clauses th))
		 :root    n
		 :mass    (mass-add (theory-mass th) dm)
		 :constraints (theory-constraints th)
		 :aux     nil 
	  ))))
  (mapcar #'build-theory (remove-if-not (lambda (i) (mass-lte (theory-mass i) max)) lst))
)))

;the following defines the function SECTIONS, which, given a family
;of sets of theories, enumerates all "selections" (picking one from each)
;which have consistent constraints and total masses adding up to less than m
(let (max-mass)
(labels 
  ((section (processed remaining constrs m) ;the main helper function
   ; (List Theory) -> (List (List Theory)) -> "{AtomicPreds} --\ {:Pos :Neg}"
     (cond       ; -> Mass -> (List Theory)
       ((null remaining) (list (make-theory   ;when the end is reached, return
	      :nodes   (apply #'append (reverse (mapcar 'theory-nodes processed)))
	      :clauses (apply #'append (reverse (mapcar 'theory-clauses processed)))
	      :mass m
	      :constraints constrs
	      :aux (mapcar 'theory-root processed))))
       ((null (car remaining)) nil)     ;no selections if some set is empty
       (t (mapcan #'copy-list          ;else, branch the given (partial) selection by every acceptable
		 (mc (i) (let ((sum  (mass-add m (theory-mass i)))  ;extension from the new set
			       (join (join constrs (theory-constraints i))))
			   (if (and (mass-lte sum (limit))
				    (not (eq join :UNDEF))) ;perhaps use (member join :UNDEF) ???
			       (section (cons i processed)
					(cdr remaining)
					join
					sum)
			       nil))
		     (car remaining)))
	  )))

   ;this simple function checks if 2 sets of constraints are consistent, and if so returns their union,
   ;which is just their join in the poset of partial functions under inclusion, else :UNDEF
   (join (c1 c2)
   ; ({AtomPreds} --\ {:Pos :Neg})^2 -> ({AtomPreds} --\ {:Pos :Neg})
     (if (null c1) c2
         (let* ((c (car c1)) (d (assoc (car c) c2)))
	   (if d
	       (if (eq (cdr c) (cdr d))
		   (join (cdr c1) c2)
		   :UNDEF)
;	       (cons c (join (cdr c1) c2)) ;this one needs :UNDEF to be member'd above
	       (join (cdr c1) (cons c c2)) ;this one's more robust, but less efficient
    ))))
   
   ;returns the maximum mass allowed, resetting if an arg is given
   (limit (&optional (new m_zero reset))
	   (if reset (setq max-mass new) max-mass)))

(defun sections (family m)  ;the main calling function
; (List (List Theory)) -> (List Theory)
(limit m)
(section nil family nil m_zero))
))
