(defmacro mc (vars form arg &rest more-args)
`(mapcar (lambda ,vars ,form) ,arg ,@more-args))

(defmacro rd (vars form init arg)
`(reduce (lambda ,vars ,form) ,arg :initial-value ,init))

(defun union* (&rest lst)
(reduce #'union lst :initial-value nil))

(defun lift-car (cons) ;AKA flatten?
(append (car cons) (cdr cons)))

(defun fint (&rest preds) ;functional intersection
(lambda (x) (every #'identity (mc (p) (funcall p x) preds))))
 
(defun makevars (n)  ;assumes n <= 10
(butlast '(v1 v2 v3 v4 v5 v6 v7 v8 v9 v0) (- 10 n)))

(defun copy-instance (i)
      (loop with i-class = (class-of i)
       with c = (allocate-instance i-class)
       for sd in (mop:class-slots i-class)
       for sn = (mop:slot-definition-name sd)
       when (slot-boundp i sn)
       do (setf (slot-value c sn)
            (slot-value i sn))
       finally (return c))) 

; given a predicate, returns a function which, given a list,
; puts the elements satisfying it into the car of a cons,
; and the rest into the cdr
(defun make-binary-sort (p)
(labels
    ((sorter (l) 
       (if (endp l) 
	   (cons nil nil)
	   (let* ((rst (sorter (cdr l)))
		  (above (car rst))
		  (below (cdr rst)))
	     (if (funcall p (car l))
		 (cons (cons (car l) above) below)
		 (cons above (cons (car l) below)))
	     ))))
  (function sorter)
))

;a home-made reducer
(defun recursor (fun lst &key init)
  (if (endp lst) 
      init
      (apply fun (list (car lst) (recursor fun (cdr lst) :init init)))
      ))

;replaces every singleton (x) by x
(defun unbra (x f)
(cond
((atom f) f)
((endp f) nil)
((equal f (list x)) x)
(t (cons (unbra x (car f))
	 (unbra x (cdr f))
))))

;circular list basics
(defun make-o (l)
  (let ((o (copy-list l)))
    (rplacd (last o) o)))
(defun o-length (o)
  (labels ((len (o s) 
	     (if (eq o s) 1 
	     (if (eq o nil) 0
		 (1+ (len (cdr o) s))))))
    (len (cdr o) o)))
(defun un-o (o)
  (labels ((cut (o s)
	     (if (eq o s) (list (car o)) (cons (car o) (cut (cdr o) s)))))
    (cut (cdr o) o)))
