;convert the given theories into the [cl-clauses] structure
(labels ((find-disjuncts (nodes clauses)
(do* ((ns nodes (cdr ns))
      (n (car ns) (car ns))
      (djs nil)
      (cls clauses)
      (rst nil))
     ((endp ns) (cons (reverse djs) (append (reverse rst) cls)))
  (case (formula-head (node-formula n))
   ((:t :f :forall) (push (make-cl-disjunct :bindlist nil :nodes (list n)) djs))
   ((:exists :and)
    (do* ((c (pop cls) (pop cls)))
	 ((equal (clause-lhs c) (list n))
	  (push (make-cl-disjunct :bindlist (formula-bindlist (node-formula n))
				  :nodes (if (or (eq (clause-op c) :exists)
						 (equal (clause-rhs c) (list n)))
;						 (= (length (clause-rhs c))
;						    (length (formula-args 
;							     (node-formula n)))))
					     (clause-rhs c)
					     (cons n (clause-rhs c))))
		djs))
      (push c rst)
      (when (endp cls)
	(format t "Error in find-disjuncts: can't find ~s~%" n)
	(break))))
   (:or    ;this happens when the disjunct is flipped
    (push (make-cl-disjunct :bindlist nil :nodes (list n)) djs))
   (t (format t "Error in find-disjuncts.~%") (break))
))))
(defun repack (clauses)
(do* ((cs (fix-exs clauses) (cdr cs))
      (c (car cs) (car cs))
      (res nil)
      (lhs (clause-lhs c) (clause-lhs c))
      (rhs (clause-rhs c) (clause-rhs c))
      (dd (domain-deficit c) (domain-deficit c))
      (djs (if (eq (clause-op c) :or) (find-disjuncts rhs cs))
	   (if (eq (clause-op c) :or) (find-disjuncts rhs cs)))
      (bl (apply #'union* (mc (x) (formula-freevars (node-formula x)) (union rhs lhs)))
	  (apply #'union* (mc (x) (formula-freevars (node-formula x)) (union rhs lhs))))
      )
     ((null cs) (nreverse res))
  (push (make-cl-clause
	 :doms (if (not (eq (clause-op c) :exists)) dd)
	 :bindlist bl
	 :lhs lhs
	 :rhs (if rhs 
	    (case (clause-op c)
	      (:forall (list (make-cl-disjunct :bindlist nil :nodes rhs)))
	      (:exists (list (make-cl-disjunct :bindlist dd  :nodes rhs)))
	      (:and (if (and (cdr rhs) lhs)
			(list (make-cl-disjunct :bindlist nil :nodes rhs))
			(mc (n) (make-cl-disjunct :bindlist nil :nodes (list n)) rhs)))
	      (:or (car djs))
	      (t (format t "Error in repack.~%")))))
	res)
  (if djs (setf cs (cdr djs)))
  (if (null (cdr cs)) (return (nreverse res)))
)))

;returns conjunctions back to the existential matrix
(defun fix-exs (cs)
(cond
((endp cs) nil)
((and (eq (clause-op (car cs)) :exists)
      (eq (formula-head (node-formula (car (clause-rhs (car cs))))) :and))
 (let* ((c (car cs))
	(n (car (clause-rhs c)))
	(cj (clause-rhs (second cs)))
	(nc (make-clause
	     :lhs (clause-lhs c)
	     :op :exists
	     :rhs (if (or (eq (length (formula-args (node-formula n))) (length cj))
			  (equal cj (list n)))
		      cj
		      (cons n cj)))))
   (cons nc (fix-exs (cddr cs)))))
(t (cons (car cs) (fix-exs (cdr cs))))
))

;sorts the clauses according to the specified order
(defun sort-clauses (th)
(labels ((nodoms (c) (null (cl-clause-doms c)))
	 (noex   (c) (or (null (cl-clause-rhs c))
			 (notany #'cl-disjunct-bindlist (cl-clause-rhs c))))
	 (nodisj (c) (<= (length (cl-clause-rhs c)) 1)))
  (let ((newtheory (copy-instance th)))
    (setf (theory-clauses newtheory)
	  (lift-car (rd (prev nextp)
			(let ((a (funcall nextp (cdr prev))))
			  (cons (append (car prev) (car a)) (cdr a)))
			(cons nil (repack (theory-clauses newtheory)))
			(mapcar #'make-binary-sort
				(list (lambda (c) (null (cl-clause-rhs c)))
				      (fint #'nodoms #'nodisj #'noex)  ;can also try to
				      (fint #'nodoms #'noex)  ;put nodoms > nodisj
				      (fint #'nodisj #'noex)
				      #'noex
				      #'nodisj
    )))))
    newtheory
)))
