(defun precl (f)
(parse-formula (cons :and (mapcar #'cdr (axiomize (tnf f))))
	       *empty-data*))

;translates the given formula, 
;and if the second argument is given, writes the output
;to a directory with that name
;accesses tptp problem name (which should be a string) and translates it,
;calling xl on the result
(defun fetch-tptp (name &optional (action :xl) fname)
(let* ((cat (subseq name 0 3))
       (formula (read-tptp-file 
	  (format nil "~~/acl/TPTP-v5.0.0/TPTP2X/lisp/~:@(~a/~a~).in" cat name))))
(case action
  (:xl (cl formula (if fname fname name)))
  (:noxl formula)
  (:display (cl formula))
  (:parse (precl formula))
  (t (cl formula action))
)))

;reads all meaningful lines in the given file and puts them into a list
(defun read-tptp-file (fname)
(with-open-file (file fname)
(do* ( (fline "a" (read-line file nil nil))
       (flag t (or (not fline)
		   (equal fline "")
		   (equal (char fline 0) #\%)))
       (res (list nil)) )
     ((not fline) 
      (unvar (add-colons (cons :and (cons (list :not (car res)) (nreverse (cdr res)))))))
     (if flag nil (let
       ((frm (read-from-string fline)))
       (case (car frm)
	 (axiom (rplacd res (cons (second frm) (cdr res))))
	 (conjecture (if (car res) (format t "%a: too much to prove...~%" fname)
			 (rplaca res (second frm)))         ) ;need this?
	 (t (format t "No mashed potatoes please!!~%"))
))))))

(defun unvar (l &optional (tag 'v_))
(cond
((atom l) l)
((endp l) l)
((eq (car l) tag) (second l))
(t (cons (unvar (car l) tag) (unvar (cdr l) tag)))
))

(defun add-colons (lst)  ;recursor is like reduce
    (recursor (lambda (x y) (subst (grammar x) x y))
	      (quote (and or eqv imp not exists forall))
	      :init lst))

(defun grammar (head)
  (case head
    (and :and)
    (or  :or)
    (imp :imp)
    (eqv :eqv)
    (not :not)
    (exists :exists)
    (forall :forall)
    (t head)
))

(defun isabellize (f)
(if (atom f) (format nil "~s" f)
(case (car f)
(:forall (format nil "(ALL~{ ~a~}. ~a)" (second f) (isabellize (third f))))
(:exists (format nil "(EX~{ ~a~}. ~a)" (second f) (isabellize (third f))))
(:and (format nil "(~{~#[~;~a~:;~a & ~]~})" (mapcar #'isabellize (cdr f))))
(:or (format nil "(~{~#[~a~:; | ~a~]~})" (mapcar #'isabellize (cdr f))))
(:imp (format nil "(~a --> ~a)" (isabellize (second f)) (isabellize (third f))))
(:not (format nil "(~~ ~a)" (isabellize (second f))))
(t (format nil "~s" f))
)))
