;creates directory with the given name and outputs each theory into it,
;each in its own separate file.
(defun write-out (ths name &optional funs (pformat :gl))
  (ensure-directories-exist (format nil "~~/acl/tests/~a/." name))
  (dotimes (i (length ths))
  (with-open-file (f (format nil "~~/acl/tests/~a/~a.~v,'0d.in" 
                             name name  
                             (ceiling (log (length ths) 10)) i)
		     :direction :output)
;    (format f "name(~a).~%~%" name)
    (dolist (j (if (eq pformat :df)
		   (print-prolog-theory (nth i ths) funs (format nil "'~a.~d'" name i))
		   (print-geolog-theory (nth i ths) funs)))
      (format f "~a" j))
    )))

;prints the theory on the screen
(defun print-out (th &optional funs)
(format t "~{~a~}" (print-geolog-theory th funs)))

;returns a list of strings representing the infile in geolog format
(defun print-prolog-theory (theory &optional funs name)
  (labels
      ((print-symb (head)
	 (case head
	   (:forall "Forall")
	   (:exists "Exists")
	   (:or     "Or" )
	   (:and    "And")
	   ((:pos :t) "t")
	   ((:neg :f) "f")
	   (t head)))
       (print-head (pred)
	 (if (not (formula-atom (node-formula pred)))
	     (format nil 
		     "~a~a_~s"
		     (print-symb (node-pole pred))
		     (print-symb (formula-head (node-formula pred)))
		     (node-index pred))
	     (format nil 
		     "~a~s"
		     (print-symb (if (eq (node-pole pred) :pos) 
				     (formula-head (node-formula pred))
				     (flip (formula-head (node-formula pred)))))
		     (atompred-head (formula-atom (node-formula pred))))
	     ))
       (print-pred (pred)
	 (format nil "~a~a"
		 (print-head pred)
		 (if (not (formula-atom (node-formula pred)))
		     (print-args (formula-freevars (node-formula pred)))
		     (print-args (mapcar #'print-term (formula-args (node-formula pred)))))
	     ))
       (print-term (term)
	 (if (term-head term)
	     (format nil 
		     "~(~s~)~@[(~{~a~#[)~:;,~]~}~]"
		     (term-head term)
		     (mapcar #'print-term (term-args term)))
	     (format nil
		     "~s"
		     (term-var term))
	     ))
       (print-conjs (conjs)
	 (format nil 
		 "~:[true~;~:*~{~a~#[~:;, ~]~}~]"
		 (mapcar #'print-pred conjs)
		 ))
       (print-disj (disj)
	 (format nil
		 "~a~a"
		 (print-doms  (cl-disjunct-bindlist disj))
		 (print-conjs (cl-disjunct-nodes disj))
		 ))
       (print-doms (vars)
	 (format nil "~{dom(~s), ~}" vars))
       (print-args (trms)
	 (format nil "~@[(~{~a~#[)~:;,~]~}~]" trms))
       (print-clauses (cs)
	 (let* ((print-clauses-counter -1)
		(fstr (format nil 
			      "~~20@<_ axiom ax~~~d,'0d~~a:~~> (~~:[~~*~~;~~a~~]~~a => ~~a).~~%~~%"
			      (ceiling (log (length cs) 10)))))
	   (mc (i) (format nil fstr
			   (incr print-clauses-counter)
			   (print-args (cl-clause-bindlist i))
			   (cl-clause-doms i)
			   (print-doms (cl-clause-doms i))
			   (print-conjs (cl-clause-lhs i))
			   (format nil "~%~26< ~>~:[false~;~{~a~#[~:;; ~]~}~]"
				   (cl-clause-rhs i)
				   (mc (j) (print-disj j) (cl-clause-rhs i))))
	       cs))))
(let ((consts (remove-if-not #'zerop funs :key #'cdr))
       (fns (remove-if #'zerop funs :key #'cdr))
       (res (print-clauses (theory-clauses theory)))
;       (node-sorter (make-binary-sort (lambda (x) (formula-atom (node-formula x)))))
       (cl-nodes (get-cl-nodes (theory-clauses theory)))
       (atomic-nodes (apply #'union* (mc (n) (if (formula-atom (node-formula n))
						 (list (formula-atom (node-formula n)))
						 nil)
					 (theory-nodes theory)))))
  (append (list	(format nil "%% Total mass: ~s~%~%" (mass-v (theory-mass theory)))
		(format nil "name(~a).~%~%" name)
;		(format nil "~@[name(~a).~%~%~]" name)
		(format nil "~:{:- dynamic ~a/~d.~%~}"
			(append (mc (x) (list (print-head x) 
				      (length (formula-freevars (node-formula x))))
				    cl-nodes)
				(mc (x) (list (format nil "t~s" (atompred-head x)) 
					      (atompred-arity x))
				    atomic-nodes)
				(mc (x) (list (format nil "f~s" (atompred-head x)) 
					      (atompred-arity x))
				    atomic-nodes)
				))
	        (format nil "~%_ axiom consts: (true => ~:[dom(g0d)~;~:*~{dom(~(~a~))~#^, ~}~]).~%~%" 
			(mapcar #'car consts))
		(if (theory-root theory)
		    (format nil "true => ~%~14< ~>~a.~%~%" (print-pred (theory-root theory)))
		    ""))
	  (mc (p) (format nil "_~30@< axiom bot~a:~> (~:*t~a, ~:*f~a => false).~%"
			  (format nil "~s~@[(~{~s~#[)~:;,~]~}~]"
				  (atompred-head p)
				  (makevars (atompred-arity p))))
	      atomic-nodes)
	  (cons (format nil "~%") res)
	  (mc (f) 
	      (format nil
		      "~:[true~;~:*~{dom(~s)~#[ ~:;, ~]~}~] => dom(~(~s~)~@[(~{~s~#[)~:;,~]~}~]).~%"
		      (makevars (cdr f))
		      (car f)
		      (makevars (cdr f)))
	      fns)
))))

;returns a list of strings representing the infile in geolog format
(defun print-geolog-theory (theory &optional funs name)
  (labels
      ((print-symb (head)
	 (case head
	   (:forall "Forall")
	   (:exists "Exists")
	   (:or     "Or" )
	   (:and    "And")
	   ((:pos :t) "t")
	   ((:neg :f) "f")
	   (t head)))
       (print-head (pred)
	 (if (not (formula-atom (node-formula pred)))
	     (format nil 
		     "~a~a_~s"
		     (print-symb (node-pole pred))
		     (print-symb (formula-head (node-formula pred)))
		     (node-index pred))
	     (format nil 
		     "~a~s"
		     (print-symb (if (eq (node-pole pred) :pos) 
				     (formula-head (node-formula pred))
				     (flip (formula-head (node-formula pred)))))
		     (atompred-head (formula-atom (node-formula pred))))
	     ))
       (print-pred (pred)
	 (format nil "~a~a"
		 (print-head pred)
		 (if (not (formula-atom (node-formula pred)))
		     (print-args (formula-freevars (node-formula pred)))
		     (print-args (mapcar #'print-term (formula-args (node-formula pred)))))
	     ))
       (print-term (term)
	 (if (term-head term)
	     (format nil 
		     "~(~s~)~@[(~{~a~#[)~:;,~]~}~]"
		     (term-head term)
		     (mapcar #'print-term (term-args term)))
	     (format nil
		     "~s"
		     (term-var term))
	     ))
       (print-conjs (conjs)
	 (format nil 
		 "~:[true~;~:*~{~a~#[~:;, ~]~}~]"
		 (mapcar #'print-pred conjs)
		 ))
       (print-disj (disj)
	 (format nil
		 "~a~a"
		 (print-doms  (cl-disjunct-bindlist disj))
		 (print-conjs (cl-disjunct-nodes disj))
		 ))
       (print-doms (vars)
	 (format nil "~{dom(~s), ~}" vars))
       (print-args (trms)
	 (format nil "~@[(~{~a~#[)~:;,~]~}~]" trms))
       (print-clauses (cs)
	 (let* ((print-clauses-counter -1)
		(fstr "~:[~*~;~a~]~a => ~a.~%~%"))
	   (mc (i) (format nil fstr
			   (cl-clause-doms i)
			   (print-doms (cl-clause-doms i))
			   (print-conjs (cl-clause-lhs i))
			   (format nil "~%~26< ~>~:[false~;~{~a~#[~:;; ~]~}~]"
				   (cl-clause-rhs i)
				   (mc (j) (print-disj j) (cl-clause-rhs i))))
	       cs))))
(let ((consts (remove-if-not #'zerop funs :key #'cdr))
       (fns (remove-if #'zerop funs :key #'cdr))
       (res (print-clauses (theory-clauses theory)))
;       (node-sorter (make-binary-sort (lambda (x) (formula-atom (node-formula x)))))
       (cl-nodes (get-cl-nodes (theory-clauses theory)))
       (atomic-nodes (apply #'union* (mc (n) (if (formula-atom (node-formula n))
						 (list (formula-atom (node-formula n)))
						 nil)
					 (theory-nodes theory)))))
  (append (list	(format nil "%% Total mass: ~s~%~%" (mass-v (theory-mass theory)))
		(if (theory-root theory)
		    (format nil "true => ~%~14< ~>~a.~%~%" (print-pred (theory-root theory)))
		    ""))
	  (mc (p) (format nil "~:*t~a, ~:*f~a => false.~%"
			  (format nil "~s~@[(~{~s~#[)~:;,~]~}~]"
				  (atompred-head p)
				  (makevars (atompred-arity p))))
	      atomic-nodes)
	  (cons (format nil "~%") res)
	  (mc (f) 
	      (format nil
		      "~:[true~;~:*~{dom(~s)~#[ ~:;, ~]~}~] => dom(~(~s~)~@[(~{~s~#[)~:;,~]~}~]).~%"
		      (makevars (cdr f))
		      (car f)
		      (makevars (cdr f)))
	      fns)
	  (list (format nil "~%true => ~:[dom(g0d)~;~:*~{dom(~(~a~))~#^, ~}~].~%~%" 
			(mapcar #'car consts)))

))))
